import pytest
from multiaddr import Multiaddr

from aioipfs.helpers import maddr_tcp_explode
from aioipfs.helpers import p2p_addr_explode
from aioipfs.helpers import peerid_reencode


class TestHelpers:
    def test_maddr_explode(self):
        assert maddr_tcp_explode(
            Multiaddr('/ip6/::1/tcp/3217')) == ('::1', 3217)

        assert maddr_tcp_explode(
            Multiaddr('/ip4/10.0.4.2/tcp/9001')) == ('10.0.4.2', 9001)

    @pytest.mark.parametrize(
        'peerid',
        ['12D3KooWD3bfmNbuuuVCYwkjnFt3ukm3qaB3hDED3peHHXawvRAi']
    )
    def test_p2p_addr_explode(self, peerid):
        assert p2p_addr_explode(
            f'/p2p/{peerid}/x/test'
        ) == (peerid, '/x/test', None)

        assert p2p_addr_explode(
            f'/p2p/{peerid}/x/hello/1.0'
        ) == (peerid, '/x/hello/1.0', '1.0')

        with pytest.raises(ValueError):
            p2p_addr_explode(f'/p2p/{peerid}')

    def test_peerid_reencode(self):
        peerid = peerid_reencode(
            '12D3KooWD3bfmNbuuuVCYwkjnFt3ukm3qaB3hDED3peHHXawvRAi',
            base='base36'
        )

        assert peerid == 'k51qzi5uqu5dhdmj65s7o239pk7w2knayz61k30js9t66wd6h82yw6iinitcuz'  # noqa



from ptpython.prompt_style import PromptStyle
from ptpython.layout import CompletionVisualisation
from prompt_toolkit import print_formatted_text, HTML
from omegaconf import DictConfig


def configure(config: DictConfig, repl) -> None:
    class CustomPrompt(PromptStyle):
        def in_prompt(self):
            return HTML("<{ipcolor}>bohort [{idx}]</{ipcolor}>: ".format(
                ipcolor=config.repl.input_prompt_color,
                idx=repl.current_statement_index
            ))

        def in2_prompt(self, width):
            return "...: ".rjust(width)

        def out_prompt(self):
            return HTML("<{opcolor}>Result[{idx}]</{opcolor}>: ".format(
                opcolor=config.repl.output_prompt_color,
                idx=repl.current_statement_index
            ))

    repl.all_prompt_styles["custom"] = CustomPrompt()
    repl.cursor_shape_config = config.repl.cursor_shape
    repl.insert_blank_line_after_output = True
    repl.prompt_style = "custom"
    repl.show_signature = config.repl.show_signature
    repl.enable_history_search = config.repl.enable_history_search
    repl.enable_auto_suggest = config.repl.enable_auto_suggest
    repl.title = 'bohort'
    repl.confirm_exit = config.repl.confirm_exit
    repl.vi_keep_last_used_mode = True
    repl.completion_menu_scroll_offset = 0
    repl.complete_while_typing = config.repl.complete_while_typing
    repl.enable_input_validation = True

    cvisual = config.repl.get('completion_visualisation')

    if cvisual in ['TOOLBAR', 'POP_UP', 'MULTI_COLUMN']:
        repl.completion_visualisation = getattr(CompletionVisualisation,
                                                cvisual)

    repl.use_code_colorscheme(config.repl.color_scheme)


def pf_text(text: str, color: str = 'ansiyellow') -> None:
    print_formatted_text(HTML(f'<{color}>{text}</{color}>'))



__version__ = '0.7.1'

from yarl import URL
from distutils.version import StrictVersion  # type: ignore
from typing import Union

import async_timeout
import asyncio
import aiohttp
import ipaddress
import re
import socket

from aiohttp import BasicAuth

from multiaddr import Multiaddr  # type: ignore
from multiaddr.exceptions import ParseError  # type: ignore
from multiaddr.exceptions import StringParseError  # type: ignore

from aioipfs import api
from aioipfs.exceptions import *  # noqa
from aioipfs.apis import dag as dag_api
from aioipfs.apis import pin as pin_api
from aioipfs.apis import multibase as multibase_api
from aioipfs.apis import p2p as p2p_api
from aioipfs.apis import pubsub as pubsub_api
from aioipfs.apis import swarm as swarm_api
from aioipfs.helpers import unusedTcpPort


RPC_API_DEFAULT_PORT = 5001


class IpfsDaemonVersion(StrictVersion):
    pass


class BearerAuth(BasicAuth):
    """
    Cheap subclass of BasicAuth to authenticate with a token
    """

    def __init__(self, token: str):
        self.token = token

    def encode(self) -> str:
        return f'Bearer {self.token}'


class AsyncIPFS(object):
    """
    Asynchronous IPFS API client

    :param str maddr: The multiaddr for the IPFS daemon's RPC API
        (this is the preferred method of specifying the node's address)
    :param str host: Hostname/IP of the IPFS daemon to connect to
    :param int port: The API port of the IPFS deamon
    :param str scheme: RPC API protocol: 'http' (default) or 'https'
    :param int conns_max: Maximum HTTP connections for this client
        (default: 0)
    :param int conns_max_per_host: Maximum per-host HTTP connections
        (default: 0)
    :param int read_timeout: Socket read timeout
    :param str api_version': IPFS protocol version ('v0' by default)
    :param bool debug: Enable client debugging traces
    :param loop: force a specific asyncio event loop
    """

    def __init__(self,
                 host: str = 'localhost',
                 port: int = RPC_API_DEFAULT_PORT,
                 scheme: str = 'http',
                 maddr: Union[Multiaddr, str] = None,
                 auth: Union[BasicAuth, BearerAuth, None] = None,
                 loop=None,
                 conns_max: int = 0,
                 conns_max_per_host: int = 0,
                 read_timeout: int = 0,
                 api_version: str = 'v0',
                 debug: bool = False):

        self._conns_max = conns_max
        self._conns_max_per_host = conns_max_per_host
        self._read_timeout = read_timeout
        self._debug = debug
        self._host, self._port, self._scheme = None, None, None
        self.loop = loop if loop else asyncio.get_event_loop()

        self._api_url = self.__compute_base_api_url(host=host,
                                                    port=port,
                                                    scheme=scheme,
                                                    maddr=maddr,
                                                    api_version=api_version)
        self.__auth = auth if isinstance(auth,
                                         (BasicAuth, BearerAuth)) else None

        self.session = self.get_session()

        # Install the API handlers
        self.core = api.CoreAPI(self)
        self.p2p = p2p_api.P2PAPI(self)
        self.block = api.BlockAPI(self)
        self.pubsub = pubsub_api.PubSubAPI(self)
        self.bitswap = api.BitswapAPI(self)
        self.bootstrap = api.BootstrapAPI(self)
        self.cid = api.CidAPI(self)
        self.config = api.ConfigAPI(self)
        self.dag = dag_api.DagAPI(self)
        self.dht = api.DhtAPI(self)
        self.diag = api.DiagAPI(self)
        self.file = api.FileAPI(self)
        self.files = api.FilesAPI(self)
        self.filestore = api.FilestoreAPI(self)
        self.key = api.KeyAPI(self)
        self.pin = pin_api.PinAPI(self)
        self.log = api.LogAPI(self)
        self.multibase = multibase_api.MultibaseAPI(self)
        self.name = api.NameAPI(self)
        self.object = api.ObjectAPI(self)
        self.refs = api.RefsAPI(self)
        self.repo = api.RepoAPI(self)
        self.routing = api.RoutingAPI(self)
        self.swarm = swarm_api.SwarmAPI(self)
        self.tar = api.TarAPI(self)
        self.stats = api.StatsAPI(self)

        self._agent_version = None

    @property
    def debug(self):
        return self._debug

    @property
    def agent_version(self):
        return self._agent_version

    @property
    def api_url(self):
        return self._api_url

    @property
    def host(self):
        return self._host

    @property
    def host_local(self):
        try:
            addr = ipaddress.ip_address(self.host)
            return addr.is_loopback
        except Exception:
            return self.host in ['localhost', socket.gethostname()]

    @property
    def port(self):
        return self._port

    @property
    def scheme(self):
        return self._scheme

    @property
    def add(self):
        return self.core.add

    @property
    def add_bytes(self):
        return self.core.add_bytes

    @property
    def add_str(self):
        return self.core.add_str

    @property
    def add_json(self):
        return self.core.add_json

    @property
    def get(self):
        return self.core.get

    @property
    def getgen(self):
        return self.core.getgen

    @property
    def id(self):
        return self.core.id

    @property
    def cat(self):
        return self.core.cat

    @property
    def dns(self):
        return self.core.dns

    @property
    def ls(self):
        return self.core.ls

    @property
    def ping(self):
        return self.core.ping

    @property
    def shutdown(self):
        return self.core.shutdown

    @property
    def version(self):
        return self.core.version

    @property
    def commands(self):
        return self.core.commands

    @property
    def auth(self):
        return self.__auth

    @auth.setter
    def auth(self, authh: Union[BasicAuth, BearerAuth, None]):
        if authh and not isinstance(authh, (BasicAuth, BearerAuth)):
            raise ValueError('Invalid authentication helper')

        self.__auth = authh

    def __compute_base_api_url(self, host=None, port=None, scheme='http',
                               maddr=None, api_version='v0'):
        """
        Compute and return the base kubo API url from the settings
        passed to the constructor. If a multiaddr is passed, use that
        to build the API URL, otherwise use host/port.

        :rtype: URL
        """

        if isinstance(maddr, str) or isinstance(maddr, Multiaddr):
            try:
                # The Multiaddr constructor can handle a string or
                # a Multiaddr instance
                node_maddr = Multiaddr(maddr)

                # Extract the host/port from the multiaddr protocols
                for proto in node_maddr.protocols():
                    if proto.name in ['ip4', 'ip6', 'dns4', 'dns6']:
                        self._host = node_maddr.value_for_protocol(
                            proto.name
                        )
                    elif proto.name == 'tcp':
                        self._port = int(
                            node_maddr.value_for_protocol(proto.name)
                        )
                    elif proto.name == 'https':
                        scheme = 'https'
                        self._scheme = scheme

                assert isinstance(self._host, str)
                assert isinstance(self._port, int)
            except (StringParseError,
                    ParseError,
                    AssertionError) as perr:
                raise InvalidNodeAddressError(
                    f'Invalid kubo node multiaddr: {maddr}: {perr}'
                )
        elif isinstance(host, str) and isinstance(port, int):
            self._host = host
            self._port = port
            self._scheme = scheme
        else:
            raise InvalidNodeAddressError()

        return URL.build(
            host=self.host,
            port=self.port,
            scheme=scheme,
            path=f'/api/{api_version}/'
        )

    def api_endpoint(self, path):
        # Returns a joined URL between the base API url and path
        return self.api_url.join(URL(path))

    async def close(self):
        await self.session.close()

    async def agent_version_get(self):
        if self.agent_version is None:
            node_idinfo = await self.core.id()

            if node_idinfo:
                agent_version = node_idinfo.get('AgentVersion')
                if not isinstance(agent_version, str):
                    raise ValueError('Invalid agent version')

                comps = agent_version.rstrip('/').split('/')

                if len(comps) >= 2:
                    self._agent_version = IpfsDaemonVersion(
                        re.sub(
                            '(-.*$)', '', comps[1]
                        )
                    )

        return self.agent_version

    async def agent_version_superioreq(self, version: str):
        """
        Returns True if the node's agent version is superior or equal to
        version

        :param str version: a kubo (formerly called go-ipfs)
            version number e.g "0.14.0"
        """
        try:
            node_vs = await self.agent_version_get()
            version_ref = IpfsDaemonVersion(version)
            return node_vs >= version_ref
        except BaseException:
            return False

    async def agent_version_post0418(self):
        return await self.agent_version_superioreq('0.4.18')

    async def agent_version_post011(self):
        return await self.agent_version_superioreq('0.11.0')

    def get_session(self, conn_timeout=60.0 * 30, read_timeout=60.0 * 10):
        return aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(
                limit=self._conns_max,
                limit_per_host=self._conns_max_per_host,
                loop=self.loop
            ), timeout=aiohttp.ClientTimeout(
                total=conn_timeout,
                sock_read=read_timeout
            )
        )

    def allocate_tcp_maddr(self) -> Multiaddr:
        """
        Find an unused TCP port on this host and return the multiaddr for it.

        :rtype: Multiaddr
        """

        port = unusedTcpPort()

        try:
            ip = ipaddress.ip_address(self.host)
            return Multiaddr(f'/ip{ip.version}/{self.host}/tcp/{port}')
        except Exception:
            # Shouldn't assume dns4 here
            return Multiaddr(f'/dns4/{self.host}/tcp/{port}')

    def timeout(self, time: float) -> async_timeout.Timeout:
        return async_timeout.timeout(time)

    def timeout_at(self, time: float) -> async_timeout.Timeout:
        return async_timeout.timeout_at(time)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()



import argparse
import asyncio
import functools
import re
import operator
import inspect

from dataclasses import dataclass
from importlib import resources
from pathlib import Path
from typing import Union, List, Dict
from ptpython import embed
from prompt_toolkit import print_formatted_text, HTML

import aioipfs
from aioipfs.apis import SubAPI
import appdirs  # type: ignore

from omegaconf import OmegaConf
from omegaconf import DictConfig, ListConfig

from .cli import configure
from .cli import pf_text


__version__ = '0.2.1'


@dataclass
class Context:
    client: aioipfs.AsyncIPFS
    interactive: bool = False


def rpc_method_config(config: DictConfig,
                      method: str) -> Union[DictConfig, None]:
    rpc = config.get('rpc_methods')

    if isinstance(rpc, DictConfig):
        return rpc.get(method)

    return None


async def _cmd_wrapper(config: DictConfig,
                       ctx: Context, method: str, *args, **kwargs):
    timeout: float = 0
    try:
        meth = operator.attrgetter(method)(ctx.client)
        assert meth

        rpc_cfg = rpc_method_config(config, method)
        if rpc_cfg:
            timeout = rpc_cfg.get('timeout', 0)
            assert isinstance(timeout, (int, float)), \
                f"Invalid timeout config for method: {method}"

            if 'defaults' in rpc_cfg:
                defaults = OmegaConf.to_container(
                    rpc_cfg.defaults)  # type: ignore

                if isinstance(defaults, dict):
                    for key, value in defaults.items():
                        if key not in kwargs and isinstance(value,
                                                            (int, float, str)):
                            kwargs[key] = value  # type: ignore

        async with ctx.client.timeout(
                timeout if timeout > 0 else None):  # type: ignore
            if inspect.isasyncgenfunction(meth):
                # async generator
                _entries: List = []

                async for entry in meth(*args, **kwargs):
                    _entries.append(entry)

                return _entries
            else:
                # coroutine

                return await meth(*args, **kwargs)
    except aioipfs.RPCAccessDenied:
        print('Access denied for this RPC endpoint! Check your credentials.')
    except (aioipfs.APIError, aioipfs.UnknownAPIError) as aerr:
        print(f'API error {aerr.code}: {aerr.message}')
    except asyncio.TimeoutError:
        pf_text(f'Timeout for method: {method} ({timeout} secs)')
    except asyncio.CancelledError as err:
        pf_text(f'Method {method} cancelled: {err}')
    except AttributeError:
        print(f'No such client method: {method}')
    except BaseException:
        raise


def get_auth_helper(creds: str) -> Union[aioipfs.BasicAuth,
                                         aioipfs.BearerAuth]:
    ma = re.match(r'^basic:(.*?):(.*?)$', creds)
    if ma:
        return aioipfs.BasicAuth(ma.group(1), ma.group(2))

    ma = re.match(r'^bearer:(.*?)$', creds)
    if ma:
        return aioipfs.BearerAuth(ma.group(1))

    raise ValueError(f'Invalid RPC credentials value: {creds}')


def save_config(cfg_path: Path, config: Union[DictConfig, ListConfig]):
    with open(cfg_path, 'wt') as f:
        OmegaConf.save(config, f)


async def start(args, cfg_dir: Path, data_dir: Path) -> None:
    cfg_path = cfg_dir.joinpath('bohort.yaml')

    if not cfg_path.exists():
        cfg_path.touch()

    with open(cfg_path, 'rt') as f:
        cfg = OmegaConf.load(f)

    with resources.files(__name__).joinpath(
            'default_config.yaml').open('r') as f:
        cfg = OmegaConf.merge(OmegaConf.load(f), cfg)

        save_config(cfg_path, cfg)

    if args.save_node:
        assert re.match(r'^[\w_-]+$', args.save_node), \
            "Invalid node name format"

        if args.save_node in cfg.nodes:
            del cfg.nodes[args.save_node]

        ncfg = OmegaConf.create({
            'nodes': {
                args.save_node: {
                    'multiaddr': args.maddr,
                    'credentials': {
                        'default': args.creds
                    }
                }
            }
        })
        cfg = OmegaConf.merge(ncfg, cfg)

        save_config(cfg_path, cfg)

    if args.node:
        node, credid = tuple(args.node.split(
            ':')) if ':' in args.node else (args.node, None)

        ncfg = cfg.nodes.get(node)
        assert ncfg, 'Node configuration does not exist!'

        maddr = ncfg.multiaddr
        creds = ncfg.credentials.get(credid if credid else 'default')
    else:
        maddr = args.maddr
        creds = args.creds

    try:
        auth = get_auth_helper(creds) if creds else None

        async with aioipfs.AsyncIPFS(maddr=maddr, auth=auth) as client:
            ires = await client.core.id()

            if not ires:
                raise Exception(
                    f'Cannot connect to kubo node with RPC: {args.maddr}')

            peer_id = ires['ID']

            print_formatted_text(HTML(
                f'<violet>bohort v{__version__} '
                f'(aioipfs v{aioipfs.__version__})</violet>'))

            print_formatted_text(HTML(
                '<ansired>Remember to "await" your calls!</ansired>')
            )

            print_formatted_text(HTML(
                f'<seagreen>{maddr} ({peer_id})</seagreen>')
            )

            ctx = Context(client=client)

            clocals: Dict = {'ctx': ctx}

            for subapi_name in [name for name, _ in inspect.getmembers(client)
                                if isinstance(_, SubAPI)]:
                coros = [
                    name for name, _ in inspect.getmembers(
                        getattr(
                            client,
                            subapi_name
                        ), inspect.isroutine) if not name.startswith('_') and
                    name not in [
                        'add_generic',
                        'add_single',
                        'fetch_json',
                        'fetch_raw',
                        'fetch_text',
                        'handle_error',
                        'decode_error',
                        'mjson_decode',
                        'post']
                ]

                for mname in coros:
                    cmd = mname if subapi_name == 'core' else \
                        f'{subapi_name}_{mname}'
                    if cmd in clocals:
                        continue

                    clocals[cmd] = functools.partial(
                        _cmd_wrapper, cfg, ctx, f'{subapi_name}.{mname}'
                    )

            await embed(
                globals={},
                locals=clocals,
                return_asyncio_coroutine=True,
                patch_stdout=True,
                configure=functools.partial(configure, cfg),
                history_filename=args.history_path if
                not args.no_history else None
            )  # type: ignore
    except aioipfs.RPCAccessDenied:
        print('RPC access denied!')
    except EOFError:
        pass
    except BaseException as err:
        raise err


def run_bohort():
    cfg_dir = Path(appdirs.user_config_dir('aioipfs')).joinpath('bohort')
    cfg_dir.mkdir(parents=True, exist_ok=True)

    data_dir = Path(appdirs.user_data_dir('aioipfs')).joinpath('bohort')
    data_dir.mkdir(parents=True, exist_ok=True)

    history_path = data_dir.joinpath('history')

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--kubo-maddr',
        '--maddr',
        '-m',
        dest='maddr',
        default='/dns4/localhost/tcp/5001',
        help="kubo RPC API multiaddr"
    )
    parser.add_argument(
        '--creds',
        '--credentials',
        '-c',
        dest='creds',
        default=None,
        help='RPC authentication credentials'
    )
    parser.add_argument(
        '--history-path',
        dest='history_path',
        default=str(history_path),
        help='History file path'
    )
    parser.add_argument(
        '--no-history',
        dest='no_history',
        action='store_true',
        default=False,
        help='Disable history'
    )
    parser.add_argument(
        '--save',
        dest='save_node',
        default=None,
        help='Save node configuration'
    )
    parser.add_argument(
        '--node',
        '--load',
        '-n',
        dest='node',
        help='Load node with this name from the config file',
        type=str
    )

    asyncio.run(start(parser.parse_args(), cfg_dir, data_dir))
